package extensions;

import java.io.IOException;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.net.Socket;
import java.net.URLDecoder;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.Map;
import java.util.TreeMap;
import java.util.concurrent.ConcurrentHashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;

import dareka.common.CloseUtil;
import dareka.common.Logger;
import dareka.common.Pair;
import dareka.extensions.Extension2;
import dareka.extensions.ExtensionManager;
import dareka.extensions.Rewriter;
import dareka.processor.HttpHeader;
import dareka.processor.HttpRequestHeader;
import dareka.processor.HttpResponseHeader;
import dareka.processor.Processor;
import dareka.processor.Resource;
import dareka.processor.StringResource;
import dareka.processor.util.GetThumbInfoUtil;


/**
 * NGIDƂĂ݂
 */
public class NGCommentExtension implements Extension2, Processor, Rewriter,
		Comparator<Map.Entry<String, Integer>> {
	
	public static final int REVISION = 110424;
	public static final String VER_STRING = "NGCommentExtension v0.5_"+ REVISION;
	public static final String LOG_PREFIX = "NGComm";
	public static final String PROP_DEBUG = "NGCommDebug";
	
	private static final Pattern MY_NGID_API_PATTERN = Pattern.compile(
			"^http://www\\.nicovideo\\.jp/cache/ng(id|stat|statxml)\\?(.+)$");
	private static final Pattern NICO_MSG_API_PATTERN = Pattern.compile(
			"^http://(msg\\.nicovideo\\.jp/\\d+/|nmsg\\.nicovideo\\.jp/)api/.*$");
	private static final Pattern NICO_NGID_API_PATTERN = Pattern.compile(
//		"^http://flapi\\.nicovideo\\.jp/api/configurengclient\\?.*source=([^&]+).*$");
			"^http://flapi\\.nicovideo\\.jp/api/configurengclient(?:\\?(.+))?$");
	
	private static final String[] PROCESSOR_SUPPORTED_METHODS =
		new String[] { HttpHeader.GET, HttpHeader.POST };
	private static final Pattern PROCESSOR_SUPPORTED_PATTERN = Pattern.compile(
			"(?:" + MY_NGID_API_PATTERN.pattern() + ")|" +
			"(?:" + NICO_NGID_API_PATTERN.pattern() + ")");
//	private static final Pattern REWRITER_SUPPORTED_PATTERN = NICO_MSG_API_PATTERN;
	private static final Pattern REWRITER_SUPPORTED_PATTERN = Pattern.compile(
			"(?:" + NICO_MSG_API_PATTERN.pattern() + ")|" +
			"(?:^http://POST/flapi\\.nicovideo\\.jp/api/configurengclient$)");
	
	private static final StringResource resultOK = ExtUtil.responseText("OK");
	private static final StringResource resultNG = ExtUtil.responseText("NG");
	
	private static volatile Map<String, Integer> listNGID = newListNGID();
	private static Map<String, Integer> newListNGID() {
		return new ConcurrentHashMap<String, Integer>();
	}
	private static final Pair<File, Long> fileInfoNGID =
			new Pair<File, Long>(new File("local/NGID.txt"), 0L);
	
	// {ȂꕨNX̂c
	private static final Map<Pattern, Object[]> listNGWORD =
			new ConcurrentHashMap<Pattern, Object[]>();
	private static final Pair<File, Long> fileInfoNGWORD =
			new Pair<File, Long>(new File("local/NGWORD.txt"), 0L);
	
	private static final LinkedList<Pair<String, Map<Integer, String[]>>> statistics =
			new LinkedList<Pair<String, Map<Integer, String[]>>>();
	
	private static ExtUtil util;
	private static Integer currentTime;
	private static int lastExpireCheck;
	private static volatile boolean saveNeeded;
	private static HttpRequestHeader lastRequestHeader;
	
	// Extension2 interface
	public void registerExtensions(ExtensionManager mgr) {
		final int REQUIRED_REVISION = 110125;
		if (ExtUtil.REVISION < REQUIRED_REVISION) {
			Logger.warning(LOG_PREFIX +
					": ERROR!!! ExtUtil_"+REQUIRED_REVISION+" OR LATER IS REQUIRED.");
		} else {
			if (util == null) {
				util = new ExtUtil(this, LOG_PREFIX, PROP_DEBUG);
				util.loadFile(fileInfoNGID, "loadNGID");
			}
			Logger.warning(LOG_PREFIX +": OeXg");
			mgr.registerProcessor(this);
			mgr.registerRewriter(this);
			updateCurrentTime();
		}
	}
	
	public String getVersionString() {
		return VER_STRING;
	}
	
	// Processor Interface
	public String[] getSupportedMethods() {
		return PROCESSOR_SUPPORTED_METHODS;
	}
	
	public Pattern getSupportedURLAsPattern() {
		return PROCESSOR_SUPPORTED_PATTERN;
	}
	
	public String getSupportedURLAsString() {
		return null;
	}
	
	public Resource onRequest(HttpRequestHeader requestHeader, Socket browser)
			throws IOException {
		String url = requestHeader.getURI();
		Matcher m;
		if ((m = MY_NGID_API_PATTERN.matcher(url)).matches()) {
			String type = m.group(1);
			String[] params = m.group(2).split("&");
			if (type.equals("id")) {
				if (params[0].equals("json")) {
					return ExtUtil.responseText(createJSON(), "application/json");
				} else if (params[0].startsWith("rm=")) {
					if (!removeNGID(params[0].substring(3))) return resultNG;
				} else {
					if (!addNGID(params[0])) return resultNG;
				}
				saveNeeded = true;
				return resultOK;
			} else if (type.startsWith("stat")) {
				boolean xml = type.endsWith("xml");
				int every = ExtUtil.getInteger("NGCommStatEvery", 0, 0);
				int last_vpos = 0;
				for (int i = 1; i < params.length; i++) {
					if (params[i].matches("every=\\d+")) {
						every = getParamInt(params[i]);
					} else if (params[i].matches("last=\\d+")) {
						last_vpos = getParamInt(params[i]) * 100 + 99;
					}
				}
				String data = getStatData(params[0], xml, every, last_vpos);
				if (data != null) {
					if (xml) {
						return ExtUtil.responseText(data, "application/xml");
					} else {
						return ExtUtil.responseText(data);
					}
				}
			}
		} else if ((m = NICO_NGID_API_PATTERN.matcher(url)).matches()) {
//			if (url.contains("type=id")) {
//				String ngid = m.group(1);
//				long delay = Long.getLong("NGCommBGDeleteNGID", 0) * 1000;
//				if (url.contains("mode=add")) {
//					addNGID(ngid);
//					if (delay > 0) setBGDeleteNGID(url, ngid, requestHeader, delay);
//				} else if (url.contains("mode=delete")) {
//					if (delay > 0) removeNGID(ngid);
//				}
//			}
			if (HttpHeader.POST.equals(requestHeader.getMethod())) {
				// POSTǂݍł܂RewriterłȂ̂ŁA
				// RewriterPOST擾ď邽߂ɕۑĂ
				lastRequestHeader = requestHeader;
			} else {
				// GET̏ꍇ͍܂œl̃^C~Oŏ
				requestHeader.setParameter("query", m.group(1));
				procNGID(requestHeader);
			}
			// IWȉĂяo
			return dareka.Main.getRewriterProcessor().onRequest(requestHeader, browser);
		}
		return StringResource.getNotFound();
	}
	
	private int getParamInt(String param) {
		return Integer.parseInt(param.split("=")[1]);
	}
	
	private static final Pattern SOURCE_PATTERN = Pattern.compile("source=([^&]+)");
	
	private void procNGID(HttpRequestHeader requestHeader) {
		String query = requestHeader.getParameter("query");
		util.debug("query %s", query);
		Matcher m;
		if (query == null || query.contains("mode=get") ||
				!(m = SOURCE_PATTERN.matcher(query)).find()) {
			return;
		}
		String ngid = m.group(1);
		long delay = Long.getLong("NGCommBGDeleteNGID", 0) * 1000;
		if (query.contains("mode=add")) {
			addNGID(ngid);
			if (delay > 0) {
				String url = requestHeader.getURI();
				if (!url.contains("?")) {
					url += "?" + query;
				}
				setBGDeleteNGID(url, ngid, requestHeader, delay);
			}
		} else if (query.contains("mode=delete")) {
			if (delay > 0) removeNGID(ngid);
		}
	}
	
	// Rewriter Interface
	public Pattern getRewriterSupportedURLAsPattern() {
		return REWRITER_SUPPORTED_PATTERN;
	}
	
	public String onMatch(Matcher match, HttpResponseHeader responseHeader, String content)
			throws IOException {
		if (match.group().startsWith("http://msg.")||match.group().startsWith("http://nmsg.")) {
			return procMatch_NICO_MSG_API(content);
		} else if (lastRequestHeader != null) {
			lastRequestHeader.setParameter("query", URLDecoder.decode(content, "UTF-8"));
			procNGID(lastRequestHeader);
		}
		return content;
	}
	
	// Comparetor Interface
	public int compare(Map.Entry<String, Integer> e1, Map.Entry<String, Integer> e2) {
		int result = e1.getValue() - e2.getValue();
		if (result == 0) {
			result = e1.getKey().compareTo(e2.getKey());
		}
		return result;
	}
	
	// ExtUtil callback
	public void onMinutes() {
		if (listNGID.isEmpty()) {
			return;
		}
		int Expire = Integer.getInteger("NGCommExpire", 30) * 3600 * 24;
		int ExpireEvery = Integer.getInteger("NGCommExpireEvery", 3600);
		boolean AnonOnly = Boolean.getBoolean("NGCommExpire184Only");
		int current_time = updateCurrentTime();
		if (Expire > 0 && current_time - lastExpireCheck > ExpireEvery) {
			ArrayList<String> expired = new ArrayList<String>();
			for (Map.Entry<String, Integer> e : listNGID.entrySet()) {
				String id = e.getKey();
				if (AnonOnly && id.matches("\\d+")) {
					continue;
				}
				int nonmatch_time = current_time - e.getValue().intValue();
				if (nonmatch_time > Expire) {
					logging("expired NGID %s %d", id, nonmatch_time);
					expired.add(id);
				}
			}
			if (expired.size() > 0) {
				for (String smid : expired) {
					listNGID.remove(smid);
				}
				lastExpireCheck = current_time;
				saveNeeded = true;
			}
		}
		onShutdown();
	}
	
	public void onShutdown() {
		if (saveNeeded) {
			saveNeeded = false;
			util.saveFile(fileInfoNGID, "saveNGID");
		}
	}
	
	private void setBGDeleteNGID(
			String url, String ngid, HttpRequestHeader requestHeader, long delay) {
		try {
			HttpRequestHeader req = new HttpRequestHeader(requestHeader.toString());
			req.setURI(url.replaceFirst("([\\?&]mode=)add", "$1delete"));
			req.setMethod(HttpHeader.GET);
			req.setParameter("NGID", ngid);
			util.<HttpRequestHeader>setSchedule("doBGDeleteNGID", req, delay);
			util.debug("NGID %s will be deleted at %,dms.", ngid, delay);
		} catch (IOException e) {
			Logger.error(e);
			util.warn("setBGDeleteNGID() failed.");
		}
	}
	
	public void doBGDeleteNGID(HttpRequestHeader requestHeader) {
		String url = requestHeader.getURI();
		String ngid = requestHeader.getParameter("NGID");
		Matcher m;
		m = NICO_NGID_API_PATTERN.matcher(url);
		boolean delete = m.matches() && url.contains("mode=delete");
		if (delete && listNGID.get(ngid) == null) {
			util.debug("NGID %s delete canceled.", ngid);
			return;
		}
		String res = ExtUtil.http_get(requestHeader, null);
		if (util.isDebug() && delete) {
			if (res.contains("<response_ngclient status=\"ok"))
				util.debug("NGID "+ngid+" was deleted from your ngclient.");
			else
				util.debug("NGID "+ngid+" delete failed, response:\n"+res);
		}
	}
	
	// O\p
	private void logging(String format, Object...args) {
		if (Boolean.getBoolean("NGCommLogging")) util.info(format, args);
	}
	
	private void logging(boolean flag, String[] s, String mes, int len) {
		if (flag) util.info("match No."+s[COMM_NO]+" "+mes+" COMM "+cutStr(s[COMMENT],len));
	}
	
	private synchronized int updateCurrentTime() {
		int time = (int)(System.currentTimeMillis() / 1000);
		currentTime = new Integer(time);
		return time;
	}
	
	private String cutStr(String str, int len) {
		int n = str.length();
		if (n == 0) return "";
		return (len < 0 || n < len) ? str : str.substring(0, len);
	}
	
	private static final Pattern VIEW_COUNTER_PATTERN = Pattern.compile(
			"<view_counter [^/>]*id=\"(\\w{2}\\d+)");
	private String getVideoId(String content) {
		Matcher m = VIEW_COUNTER_PATTERN.matcher(content);
		return m.find() ? m.group(1) : null;
	}
	
	private static final Pattern MSG_CHAT_PATTERN = Pattern.compile(
			"<chat( .+?\")(?:>([^<]*)</chat>|\\s*/>)");
	private static final int COMM_NO = 1;
	private static final int VPOS    = 2;
	private static final int MAIL    = 3;
	private static final int USER_ID = 4;
	private static final int COMMENT = 5;
	private static final int _RESULT = 6; // KŌɂ邱
	
	private static final String RESULT_NORMAL  = "NORMAL";
	private static final String RESULT_OWNER   = "OWNER";
	private static final String RESULT_DELETED = "DELETED";
	private static final String RESULT_NGMATCH = "NGMATCH";
	
	private String[] createChatInfo(Matcher m) {
		String[] s = new String[_RESULT + 1];
		s[0] = m.group(1);
		parseChat(s);
		if (s[VPOS].length() > 0 && !s[VPOS].matches("\\d+")) {
			s[VPOS] = "0";
			util.warn("invalid vpos, force vpos=0.\n%s", m.group());
		}
		s[COMMENT] = m.groupCount() > 1 && m.group(2) != null ? m.group(2) : "";
		s[_RESULT] = RESULT_NORMAL;
		return s;
	}
	
	private void parseChat(String[] s) {
		assert s.length == _RESULT + 1;
		s[COMM_NO] = parseChatAttr(s[0], "no");
		s[VPOS   ] = parseChatAttr(s[0], "vpos");
		s[MAIL   ] = parseChatAttr(s[0], "mail");
		s[USER_ID] = parseChatAttr(s[0], "user_id");
	}
	
	private String parseChatAttr(String attrs, String attrName) {
		int index = attrs.indexOf(attrName.concat("=\""));
		if (index > 0) {
			if (attrs.charAt(index - 1) == 0x20) {
				int start = index + attrName.length() + 2;
				return attrs.substring(start, attrs.indexOf("\"", start));
			}
		}
		return "";
	}
	
	private boolean isOwnerComment(String[] s) {
		return s[0].contains(" fork=\"1");
	}
	
	private static final Pattern THREAD_RESULT_PATTERN = Pattern.compile(
			"<thread [^>]*(?<= )thread=\"(\\d+)\"[^>]*>");
	private static final Pattern VIDEO_ID_PATTERN = Pattern.compile(
			"<video_id>([a-z]{2}\\d+)</video_id>");
	private static final Pattern USER_ID_PATTERN = Pattern.compile(
			"<user_id>(\\d+)</user_id>");
	private static final Pattern VIDEO_LENGTH_PATTERN = Pattern.compile(
			"<length>(\\d+):(\\d+)</length>");
	private static final File original_xml = new File("NGComm_original.xml");
	private static final File modified_xml = new File("NGComm_modified.xml");
	
	private String procMatch_NICO_MSG_API(String content) {
		String videoId = null, threadId = null;
		matchInit();
		Matcher m = THREAD_RESULT_PATTERN.matcher(content);
		if(m.find()) {
			boolean mymemory = m.group().contains("res_offset=");
			if (!mymemory || !LatestOnly) {
				if ((videoId = getVideoId(content)) == null && !LatestOnly)
					threadId = m.group(1);
			}
		}
		if (LatestOnly && videoId == null && threadId == null) {
			util.debug("latest comment only, skip.");
			return content;
		}
		String thumbinfo = null;
		if (NeedsThumbInfo || ShowThumbInfo) {
			thumbinfo = getThumbInfo(videoId, threadId);
		}
		if (thumbinfo != null && videoId == null) {
			m = VIDEO_ID_PATTERN.matcher(thumbinfo);
			if(m.find()) videoId = m.group(1);
		}
		if (util.isDebug() && content.length() > 1024) {
			writeTo(original_xml, content.replaceAll("<[a-z]","\n$0"));
		}
		StringBuffer sb = new StringBuffer(content.length());
		Map<Integer, String[]> statMap = getStatMap(videoId, true);
		m = MSG_CHAT_PATTERN.matcher(content);
		while (m.find()) {
			String[] s = createChatInfo(m);
			boolean matched = false;
			if (!s[0].contains(" deleted=") && s[COMMENT].length() > 0) {
				if (isOwnerComment(s)) {
					matched = matchNGJUMP(s, thumbinfo);
					s[_RESULT] = RESULT_OWNER;
				} else {
					matched = matchNGCMD(s, NGCommand);
					if (!matched) {
						matched = matchNGID(s);
						if (!matched) {
							matched = matchNGWORD(s, videoId, thumbinfo);
							if(matched && AutoNGID) addNGID(s[USER_ID]);
						}
					}
					if(matched) s[_RESULT] = RESULT_NGMATCH;
				}
			} else {
				s[_RESULT] = RESULT_DELETED;
			}
			if (statMap != null && s[_RESULT] != RESULT_OWNER) {
				statMap.put(new Integer(s[COMM_NO]), s);
			}
			m.appendReplacement(sb, matchReplace(matched, s));
		}
		putStatMap(videoId, statMap);
		m.appendTail(sb);
		String modified = sb.toString();
		if (util.isDebug() && content.length() > 1024) {
			writeTo(modified_xml, modified.replaceAll("<[a-z]","\n$0"));
		}
		return modified;
	}
	
	private String getThumbInfo(String videoId, String threadId) {
		String thumbinfo = null;
		try {
			// LbVɂ\̍videoIdDIɎ擾
			thumbinfo = GetThumbInfoUtil.get(
					videoId != null ? videoId : threadId);
			if (thumbinfo == null) {
				util.warn("getthumbinfo failed.");
			}
			if (ShowThumbInfo) {
				util.info("ShowThumbInfo["+videoId+"]\n"+thumbinfo);
			}
		} catch (IOException e) {
			if(util.isDebug()) Logger.error(e);
		}
		return thumbinfo;
	}
	
	private Map<Integer, String[]> getStatMap(String smid, boolean create) {
		Map<Integer, String[]> statMap = null;
		synchronized (statistics) {
			for (int i = 0; i < statistics.size(); i++) {
				if (statistics.get(i).first.equals(smid)) {
					Pair<String, Map<Integer, String[]>> e = statistics.remove(i);
					statMap = e.second;
					statistics.add(e); break;
				}
			}
			int max = ExtUtil.getInteger("NGCommStatMax", 5, 0);
			if (statMap == null && create && smid != null && max > 0) {
				statMap = new ConcurrentHashMap<Integer, String[]>();
			}
			while(statistics.size() > max) {
				String removed = statistics.remove().first;
				util.debug("statMap<%s> removed, remaining=%d.",
						removed, statistics.size());
			}
			if (util.isDebug() && statistics.size() > 0) {
				StringBuilder sb = new StringBuilder("statMap");
				for (Pair<String, Map<Integer, String[]>> p : statistics)
					sb.append(" ").append(p.first);
				util.debug(sb.toString());
			}
		}
		return statMap;
	}
	
	private void putStatMap(String smid, Map<Integer, String[]> statMap) {
		if (statMap != null/* && statMap.size() > 0*/) {
			synchronized (statistics) {
				Pair<String, Map<Integer, String[]>> e = null;
				for (int i = 0; i < statistics.size(); i++) {
					if (statistics.get(i).first.equals(smid)) {
						e = statistics.remove(i); break;
					}
				}
				if (e != null) {
					e.second = statMap;
				} else {
					e = new Pair<String, Map<Integer, String[]>>(smid, statMap);
				}
				statistics.add(e);
				util.debug("statMap<%s> added, size=%d.", e.first, e.second.size());
			}
		}
	}
	
	private String getStatData(String smid, boolean xml, int every, int last_vpos) {
		Map<Integer, String[]> statMap = getStatMap(smid, false);
		if (statMap != null) {
			TreeMap<Integer, int[]> data = new TreeMap<Integer, int[]>();
			every = initStatData(statMap, data, smid, every, last_vpos);
			for (Map.Entry<Integer, String[]> e : statMap.entrySet()) {
				String[] s = e.getValue();
				Integer key = getStatDataKey(s[VPOS], every, last_vpos);
				if(key == null) continue;
				int[] d = data.get(key);
				if (d != null) {
					d[0]++; // Comments
					if(s[_RESULT] == RESULT_NGMATCH) d[1]++;
					if(s[_RESULT] == RESULT_DELETED) d[2]++;
					data.put(key, d);
				} else {
					// L蓾Ȃ͂c
					util.warn("stat data '%s' not found.", key);
				}
			}
			util.debug("statMap<%s> returned, size=%d.", smid, statMap.size());
			if (xml) {
				return makeStatDataXML(data);
			} else {
				StringBuilder sb = new StringBuilder();
				for (Map.Entry<Integer, int[]> e : data.entrySet()) {
					sb.append(getTimeString(e.getKey())).append(',');
					int[] d = e.getValue();
					sb.append(d[0]).append(',');
					sb.append(d[1]).append(',');
					sb.append(d[2]).append('\n');
				}
				return sb.toString();
			}
		}
		return null;
	}
	
	private int initStatData(Map<Integer, String[]> statMap,
			TreeMap<Integer, int[]> data, String smid, int every, int last_vpos) {
		int max_vpos = last_vpos;
		if (max_vpos == 0) {
			for (String[] s : statMap.values()) {
				int vpos = Integer.parseInt(s[VPOS]);
				if(max_vpos < vpos) max_vpos = vpos;
			}
			int length = getVideoLength(smid) * 100 + 99;
			if(max_vpos < length) max_vpos = length;
		}
		if (every == 0) { // 0Ȃ玩
			if (max_vpos < 15000) {      // 230b
				every = 1;
			} else if (max_vpos < 60000) {  // 10
				every = 3;
			} else if (max_vpos < 180000) { // 30
				every = 5;
			} else {                    // 30ȏ
				every = 10;
			}
		}
		for (int vpos = 0; max_vpos >= vpos; vpos += every * 100) {
			data.put(vpos / 100, new int[3]);
		}
		return every;
	}
	
	private int getVideoLength(String smid) {
		String thumbinfo = getThumbInfo(smid, null);
		if (thumbinfo != null) {
			Matcher m = VIDEO_LENGTH_PATTERN.matcher(thumbinfo);
			if (m.find()) {
				int min = Integer.parseInt(m.group(1));
				int sec = Integer.parseInt(m.group(2));
				return min * 60 + sec;
			}
		}
		return 0;
	}
	
	private Integer getStatDataKey(String str_vpos, int every, int last_vpos) {
		int vpos = Integer.parseInt(str_vpos);
		if (vpos <= last_vpos) {
			return new Integer((vpos / (every * 100)) * every);
		}
		return null;
	}
	
	private String getTimeString(int sec) {
		return String.format("%d:%02d", sec / 60, sec % 60);
	}
	
	private String makeStatDataXML(TreeMap<Integer, int[]> data) {
		final String[] titles = new String[] {
				"Comments", "Matches", "Deleted" };
		StringBuilder sb = new StringBuilder(
				  "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n" +
				  "<chart>\n" +
				  "\t<series>\n");
		int n = 0;
		for (Integer key : data.keySet()) {
			sb.append("\t\t<value xid=\"").append(n++).append("\">");
			sb.append(getTimeString(key)).append("</value>\n");
		}
		sb.append("\t</series>\n");
		sb.append("\t<graphs>\n");
		for (int i = 0; i < 3; i++) {
			sb.append("\t\t<graph gid=\"").append(i);
			sb.append("\" title=\"").append(titles[i]).append("\">\n");
			n = 0;
			for (int[] d : data.values()) {
				sb.append("\t\t\t<value xid=\"").append(n++).append("\">");
				sb.append(d[i]).append("</value>\n");
			}
			sb.append("\t\t</graph>\n");
		}
		sb.append("\t</graphs>\n");
		sb.append("</chart>\n");
		return sb.toString();
		
	}
	
	// ݊ێ̂߂ɎcĂ
	private boolean NGCommLoggingNGID   = false;
	private boolean NGCommLoggingNGWORD = false;
	private boolean NGCommAddUserID     = false;
	
	// {IɃ[v̏Ȃ̂ŃtOn͑Să[vOŎ擾
	private boolean LoggingNGID, LoggingNGWORD, ShowUserID;
	private boolean Logging, AutoNGID, ShowMatched, KillJUMP, KillJUMP_ALL;
	private boolean NeedsThumbInfo, ShowThumbInfo, LatestOnly;
	private Pattern NGCommand;
	
	private void matchInit() {
		LoggingNGID   = NGCommLoggingNGID   || Boolean.getBoolean("NGCommLoggingNGID");
		LoggingNGWORD = NGCommLoggingNGWORD || Boolean.getBoolean("NGCommLoggingNGWORD");
		ShowUserID    = NGCommAddUserID     || Boolean.getBoolean("NGCommShowUserID");
		Logging       = Boolean.getBoolean("NGCommLogging");
		AutoNGID      = Boolean.getBoolean("NGCommAutoNGID");
		ShowMatched   = Boolean.getBoolean("NGCommShowMatched");
		ShowThumbInfo = Boolean.getBoolean("NGCommShowThumbInfo");
		LatestOnly    = Boolean.getBoolean("NGCommLatestOnly");
		
		String prop = System.getProperty("NGCommKillJUMP", "false");
		KillJUMP      = !"false".equals(prop);
		KillJUMP_ALL  =  "all".equals(prop);
		
		prop = System.getProperty("NGCommNGCMD", "");
		try {
			NGCommand = prop.length() == 0 ? null : Pattern.compile(prop);
		} catch (PatternSyntaxException e) {
			util.warn("regex error in NGCommNGCMD '%s'.", prop);
		}
		updateCurrentTime();
		if (listNGID.isEmpty() && !saveNeeded) {
			// TODO r䂪Kv
			fileInfoNGID.second = 0L;  // Ôߋ̏ꍇ͂x[h
		}
		util.loadFile(fileInfoNGID,   "loadNGID");
		util.loadFile(fileInfoNGWORD, "loadNGWORD");
	}
	
	private String matchReplace(boolean matched, String[] s) {
		String prefix = "", attrs = s[0], comment = s[COMMENT];
		if (ShowUserID && s[USER_ID].length() > 0)
			prefix = "[" + cutStr(s[USER_ID] ,8) + "] ";
		if (!matched && prefix.length() == 0)
			return "$0"; // NOT REPLACE
		if (matched) {
			if (ShowMatched) {
				if (isOwnerComment(s)) {
					String mail = " mail=\"@30 shita cyan\"";
					attrs = attrs.contains(" mail=") ?
						attrs.replaceFirst(" mail=\"[^\"]*\"", mail) :
						attrs.replaceFirst(" date=\"[^\"]+\"", "$0".concat(mail));
					prefix = "[RNG] ".concat(prefix);
				} else {
					prefix = "[NG] ".concat(prefix);
				}
			} else {
				attrs = attrs.replaceFirst(
						" (user_id|fork)=\"[^\"]+\"", "$0 deleted=\"1\"");
				prefix = comment = "";
			}
		}
		return "<chat" + attrs + ">" + prefix + comment + "</chat>";
	}
	
	private boolean matchNGCMD(String[] s, Pattern pat) {
		if (pat != null && s[MAIL].length() > 0) {
			Matcher m = pat.matcher(s[MAIL]);
			if (m.find()) {
				logging(Logging, s, "NGCMD ".concat(m.group()), 15);
				return true;
			}
		}
		return false;
	}
	
	private boolean matchNGID(String[] s) {
		if (listNGID.containsKey(s[USER_ID])) {
			logging(LoggingNGID, s, "NGID ".concat(s[USER_ID]), 15);
			listNGID.put(s[USER_ID], currentTime);
			return true;
		}
		return false;
	}
	
	private boolean matchNGWORD(String[] s, String videoId, String thumbinfo) {
		int vpos = Integer.parseInt(s[VPOS]);
		for (Map.Entry<Pattern, Object[]> e : listNGWORD.entrySet()) {
			Object[] params = e.getValue();
			int[] range = (int[])params[PARAM_RANGE];
			if (vpos < range[0] || range[1] < vpos) {
				continue;
			}
			boolean doMatch = !(Boolean)params[PARAM_NOT];
			String ids[] = (String[])params[PARAM_IDS];
			Pattern patThumb = (Pattern)params[PARAM_THUMB];
			if (ids != null) {
				boolean matched = false;
				for (String id : ids) {
					if (id.equals(videoId)) {
						matched = true; break;
					}
				}
				if(!matched) doMatch = !doMatch;
			} else if (patThumb != null) {
				if (thumbinfo != null) {
					if (!patThumb.matcher(thumbinfo).find())
						doMatch = !doMatch;
				} else {
					// thumbinfoꍇ͈Ŝ߃}b`Ȃ
					doMatch = false;
				}
			}
			if (doMatch) {
				Pattern pat = e.getKey();
				if ((Boolean)params[PARAM_NGCMD]) {
					if (matchNGCMD(s, pat))
						return true;
				} else { // NGWORD
					Matcher m = pat.matcher(s[COMMENT]);
					if (m.find()) {
						logging(LoggingNGWORD, s,
								"NGWORD "+m.pattern()+" vpos="+vpos, -1);
						return true;
					}
				}
			}
		}
		return false;
	}
	
	private static final Pattern SCRIPT_JUMP_PATTERN = Pattern.compile(
		"^(?:[@](?:|L[[h)Wv[ @].*|/.*jump\\(['\"])([a-z]{2}[0-9]+)",
		Pattern.DOTALL);
	private boolean matchNGJUMP(String[] s, String thumbinfo) {
		if (KillJUMP) {
			Matcher m = SCRIPT_JUMP_PATTERN.matcher(s[COMMENT]);
			if (m.lookingAt()) {
				String smid = m.group(1);
				boolean ngjump = KillJUMP_ALL ||
						ExtUtil.isNGvideoId(smid)|| ExtUtil.hasRemoved(smid);
				if (!ngjump && thumbinfo != null &&
						(m = USER_ID_PATTERN.matcher(thumbinfo)).find())
					ngjump = ExtUtil.isNGuserId(m.group(1));
				if (ngjump) {
					logging(Logging, s, "NGJUMP ".concat(smid), -1);
					return true;
				}
			}
		}
		return false;
	}
	
	private boolean addNGID(String ngid) {
		if (ngid.matches("^(\\d+|[^\\s]{27})$") || ngid.length() == 27) {
			util.info("add NGID " + ngid + " " + currentTime);
			listNGID.put(ngid, currentTime);
			saveNeeded = true;
			return true;
		} else {
			util.info("invalid NGID " + ngid);
		}
		return false;
	}
	
	private boolean removeNGID(String ngid) {
		Integer time = listNGID.remove(ngid);
		if (time != null) {
			util.info("remove NGID " + ngid + " " + time);
			return true;
		}
		return false;
	}
	
	private static final Pattern NGID_PATTERN = Pattern.compile(
			"^(\\d+|[^#][^\\s]{26})(?:\\s+(\\d+))?$"); // NGID[<SP>LAST_TIME]
	public void loadNGID(BufferedReader br) throws IOException {
		if (br != null) {
			Map<String, Integer> loaded = newListNGID();
			String line;
			while ((line = br.readLine()) != null) {
				Matcher m = NGID_PATTERN.matcher(line);
				if (m.matches()) {
					loaded.put(m.group(1), (m.group(2) != null) ?
							new Integer(m.group(2)) : currentTime);
				}
			}
			listNGID = loaded;
		} else {
			listNGID.clear();
		}
	}
	
	public void saveNGID(BufferedWriter bw) throws IOException {
		ArrayList<Map.Entry<String, Integer>> list =
			new ArrayList<Map.Entry<String, Integer>>(listNGID.entrySet());
		Collections.sort(list, this);
		for (Map.Entry<String, Integer> e : list) {
			bw.write(e.getKey());
			bw.write(" ");
			bw.write(e.getValue().toString());
			bw.write("\r\n");
		}
	}
	
	public void loadNGWORD(BufferedReader br) throws IOException {
		NGCommLoggingNGID   = false;
		NGCommLoggingNGWORD = false;
		NGCommAddUserID     = false;
		NeedsThumbInfo      = false;
		
		listNGWORD.clear();
		if (br != null) {
			logging("loading '%s'", fileInfoNGWORD.first);
			String line, kind;
			for (int num = 1; (line = br.readLine()) != null; num++) {
				if (line.length() == 0 || line.startsWith("#")) {
					continue;
				} else if (line.equals("NGCommLoggingNGID")) {
					NGCommLoggingNGID = true;
				} else if (line.equals("NGCommLoggingNGWORD")) {
					NGCommLoggingNGWORD = true;
				} else if (line.equals("NGCommAddUserID")) {
					NGCommAddUserID = true;
				} else {
					if((kind = parseNGWORD(line)) != null) {
						util.warn("%s error in %s, line %d.",
								kind, fileInfoNGWORD.first, num);
					}
				}
			}
		}
	}
	
	private static final int PARAM_RANGE = 0;
	private static final int PARAM_IDS   = 1;
	private static final int PARAM_NOT   = 2;
	private static final int PARAM_NGCMD = 3;
	private static final int PARAM_THUMB = 4;
	
	private static final Pattern NGWORD_LINE_PATTERN = Pattern.compile(
			"^(?:([a-z]+)=(!?[^'\\s]+|'[^']+')\\s+)?([^\\s].*)$");
	private String parseNGWORD(String regex) {
		 Matcher m = NGWORD_LINE_PATTERN.matcher(regex);
		if (!m.find()) {
			return "format";
		}
		int[] range = { 0, Integer.MAX_VALUE };
		String[] ids = null;
		boolean notMatch = false, ngcmd = false;
		String thumbRegex = null;
		try {
			if (m.group(1) != null) {
				String key = m.group(1), value = m.group(2);
				if (value.startsWith("!")) {
					value = value.substring(1);
					notMatch = true;
				}
				if (value.startsWith("'")) {
					value = value.substring(1, value.length() - 1);
				}
				if ("vpos".equals(key) || "sec".equals(key)) {
					parseRange(value, range);
					if ("sec".equals(key)) {
						// vpos̒Pʂ10~bȂ̂100{
						range[0] *= 100; range[1] = range[1]*100 + 99;
					}
				} else if ("id".equals(key)) {
					ids = value.split(",");
				} else if ("thumbinfo".equals(key)) {
					thumbRegex = value;
				} else {
					return "keyword";
				}
			}
			if (m.group(3).indexOf("NGCMD:") == 0) {
				regex = m.group(3).substring(6);
				ngcmd = true;
			} else {
				regex = m.group(3);
			}
		} catch (NumberFormatException e) {
			return "range";
		}
		try {
			Pattern pat = Pattern.compile(regex);
			Object[] params = {
				range, ids, new Boolean(notMatch), new Boolean(ngcmd),
				thumbRegex != null ? Pattern.compile(thumbRegex) : null };
			listNGWORD.put(pat, params);
			logging("loaded NGWORD %s", m.group());
			if (util.isDebug()) {
				StringBuilder sb = new StringBuilder();
				sb.append("range=" + range[0] + "-" + range[1]);
				if (ids != null) {
					sb.append(" ids=");
					for(int i=0; i<ids.length; i++) sb.append(ids[i] + ",");
				}
				sb.append(" not=" + notMatch + " ngcmd=" + ngcmd);
				sb.append(" thumb=" + thumbRegex + " regex=" + regex);
				util.debug(sb.toString());
			}
		} catch (PatternSyntaxException e) {
			return "regex";
		}
		if(thumbRegex != null) NeedsThumbInfo = true;
		return null; // SUCCESS
	}
	
	private void parseRange(String params, int[] range) {
		String[] s = params.split("-");
		range[0] = Integer.parseInt(s[0]);
		if (s.length > 1) {
			range[1] = Integer.parseInt(s[1]);
		} else {
			range[1] = range[0];
		}
	}
	
	private String createJSON() {
		// 1ső40Ƃďobt@m
		StringBuilder sb = new StringBuilder((listNGID.size() + 1) * 40);
		sb.append("{\r\n");
		if (listNGID.size() > 0) {
			for (Map.Entry<String, Integer> e : listNGID.entrySet())
				sb.append("\"" + e.getKey() + "\": " + e.getValue().toString() + ",\r\n");
			sb.deleteCharAt(sb.lastIndexOf(","));
		}
		sb.append("}");
		return sb.toString();
	}
	
	private boolean writeTo(File file, String content) {
		synchronized (file) {
			BufferedWriter bw = null;
			try {
				bw = new BufferedWriter(new FileWriter(file));
				bw.write(content);
				return true;
			} catch (IOException e) {
				Logger.error(e);
			} finally {
				CloseUtil.close(bw);
			}
		}
		return false;
	}
	
}
